"""Utility functions"""
import argparse
import dataclasses
import functools
import re
from collections import defaultdict
from dataclasses import dataclass, field, MISSING, Field
from typing import *
from enum import Enum
import logging
import builtins
import json

logger = logging.getLogger(__name__)

builtin_types = [getattr(builtins, d) for d in dir(builtins) if isinstance(getattr(builtins, d), type)]

T = TypeVar("T")
Dataclass = TypeVar("Dataclass")
DataclassType = Type[Dataclass]

def MutableField(default: T, init: bool = True, repr: bool = True, hash: bool = None, compare: bool = True, metadata: Dict[str, Any] = None) -> T:
    return field(default_factory=lambda: default, init=init, repr=repr, hash=hash, compare=compare, metadata=metadata) 

class InconsistentArgumentError(RuntimeError):
    """
    Error raised when the number of arguments provided is inconsistent when parsing multiple instances from command line.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.MetavarTypeHelpFormatter):
    """Little shorthand for using both of argparse's ArgumentDefaultHelpFormatter and MetavarTypeHelpFormatter classes.
    """
    pass

def camel_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

TRUE_STRINGS: List[str] = ['yes', 'true', 't', 'y', '1']
FALSE_STRINGS: List[str] = ['no', 'false', 'f', 'n', '0']

def str2bool(v: str) -> bool:
    """
    Taken from https://stackoverflow.com/questions/15008758/parsing-boolean-values-with-argparse
    """
    if isinstance(v, bool):
        return v
    v = v.strip()
    if v.lower() in TRUE_STRINGS:
        return True
    elif v.lower() in FALSE_STRINGS:
        return False
    else:
        raise argparse.ArgumentTypeError(f"Boolean value expected for argument, received '{v}'")


def get_item_type(container_type: Type[Container[T]]) -> T:
    """Returns the `type` of the items in the provided container `type`. When no type annotation is found, or no item type is found, returns `typing.Type[typing.Any]`.
    Note, if a Tuple container type is passed, only the first argument to the type is actually used.
    >>> import typing
    >>> from typing import List, Tuple
    >>> get_item_type(list)
    typing.Type[typing.Any]
    >>> get_item_type(List)
    typing.Type[typing.Any]
    >>> get_item_type(tuple)
    typing.Type[typing.Any]
    >>> get_item_type(Tuple)
    typing.Type[typing.Any]
    >>> get_item_type(List[int])
    <class 'int'>
    >>> get_item_type(List[str])
    <class 'str'>
    >>> get_item_type(List[float])
    <class 'float'>
    >>> get_item_type(List[float])
    <class 'float'>
    >>> get_item_type(List[Tuple])
    typing.Tuple
    >>> get_item_type(List[Tuple[int, int]])
    typing.Tuple[int, int]
    >>> get_item_type(Tuple[int, str])
    <class 'int'>
    >>> get_item_type(Tuple[str, int])
    <class 'str'>
    >>> get_item_type(Tuple[str, str, str, str])
    <class 'str'>

    Arguments:
        list_type {Type} -- A type, preferably one from the Typing module (List, Tuple, etc).
    
    Returns:
        Type -- the type of the container's items, if found, else Any.
    """
    if container_type in {list, tuple}:
        # the built-in `list` and `tuple` types don't have annotations for their item types.
        return Type[Any]
    type_arguments = getattr(container_type, "__args__", [str])
    if type_arguments:
        return type_arguments[0]
    else:
        return Type[Any]

def get_argparse_type_for_container(container_type: Type) -> Type:
    """Gets the argparse 'type' option to be used for a given container type.
    When an annotation is present, the 'type' option of argparse is set to that type.
    if not, then the default value of 'str' is returned.
    
    Arguments:
        container_type {Type} -- A container type (ideally a typing.Type such as List, Tuple, along with an item annotation: List[str], Tuple[int, int], etc.)
    
    Returns:
        typing.Type -- the type that should be used in argparse 'type' argument option.
    """
    T = get_item_type(container_type)
    return T if T is not Type[Any] else str



def _mro(t: Type) -> List[Type]:
    return getattr(t, "mro", lambda: [])()

def is_subtype_of(some_type: Type):
    def func(field: Field) -> bool:
        return some_type in _mro(field.type)
    return  func


def is_list(t: Type) -> bool:
    return list in _mro(t)

def is_tuple(t: Type) -> bool:
    return tuple in _mro(t)

def is_enum(t: Type) -> bool:
    return Enum in _mro(t)

def is_bool(t: Type) -> bool:
    return bool in _mro(t)

def is_tuple_or_list(t: Type) -> bool:
    return is_list(t) or is_tuple(t)

def is_tuple_or_list_of_dataclasses(t: Type) -> bool:
    return is_tuple_or_list(t) and dataclasses.is_dataclass(get_item_type(t))

def _parse_multiple_containers(tuple_or_list: type, append_action: bool = False) -> Callable[[str], List[Any]]:
    T = get_argparse_type_for_container(tuple_or_list)
    factory = tuple if is_tuple(tuple_or_list) else list
    
    result = factory()

    def parse_fn(v: str) -> List[Any]:
        # nonlocal result
        # print(f"Parsing a {tuple_or_list} of {T}s, value is: {v}, type is {type(v)}")
        v = v.strip()
        if v.startswith("[") and v.endswith("]"):
            v = v[1:-1]
        
        separator = " "
        for sep in [","]: # TODO: maybe add support for other separators?
            if sep in v:
                separator = sep
        str_values = [v.strip() for v in v.split(separator)]
        T_values = [T(v_str) for v_str in str_values]
        values = factory(v for v in T_values)
        # print("values:", values)
        if append_action:
            result += values
            return result
        else:
            return values
    return parse_fn


def _parse_container(tuple_or_list: type,) -> Callable[[str], List[Any]]:
    T = get_argparse_type_for_container(tuple_or_list)
    factory = tuple if is_tuple(tuple_or_list) else list

    def parse_fn(v: str) -> List[Any]:
        # TODO: maybe we could use the fact we know this isn't for a list of lists to make this better somehow.
        logger.debug(f"Parsing a {tuple_or_list} of {T}s, value is: '{v}', type is {type(v)}")
        v = v.strip()
        if v.startswith("[") and v.endswith("]"):
            v = v[1:-1]
        
        separator = " "
        for sep in [","]: # TODO: maybe add support for other separators?
            if sep in v:
                separator = sep
        str_values = [v.strip() for v in v.split(separator)]
        T_values = [T(v_str) for v_str in str_values]
        values = factory(v for v in T_values)
        logger.debug(f"returning values: {values}")
        return values

    return parse_fn

def setattr_recursive(obj: object, attribute_name: str, value: Any):
    if "." not in attribute_name:
        setattr(obj, attribute_name, value)
    else:
        parts = attribute_name.split(".")
        child_object = getattr(obj, parts[0])
        setattr_recursive(child_object, ".".join(parts[1:]), value)


def parent_and_child(destination: str) -> Tuple[str, str]:
    splits = destination.split(".")
    parent = ".".join(splits[:-1])
    attribute_in_parent = splits[-1]
    return parent, attribute_in_parent


def get_nesting_level(possibly_nested_list):
    if not isinstance(possibly_nested_list, (list, tuple)):
        return 0
    elif len(possibly_nested_list) == 0:
        return 1
    else:
        return 1 + max(
            get_nesting_level(item) for item in possibly_nested_list
        )



def default_value(field: dataclasses.Field) -> Optional[Any]:
    """Returns the default value of a dataclass field, if available.

    Args:
        field (dataclasses.Field[T]): The dataclasses.Field to get the default value of.

    Returns:
        Optional[T]: The default value for that field, if present, or None otherwise.
    """

    if field.default is not dataclasses.MISSING:
        return field.default
    elif field.default_factory is not dataclasses.MISSING:  # type: ignore
        return field.default_factory()  # type: ignore
    else:
        return None


class JsonSerializable:
    """
    Enables reading and writing a Dataclass to a JSON file.
    
    >>> from dataclasses import dataclass
    >>> from simple_parsing.utils import JsonSerializable
    >>> @dataclass
    ... class Config(JsonSerializable):
    ...   a: int = 123
    ...   b: str = "456"
    ... 
    >>> config = Config()
    >>> config
    Config(a=123, b='456')
    >>> config.save_json("config.json")
    >>> config_ = Config.load_json("config.json")
    >>> config_
    Config(a=123, b='456')
    >>> assert config == config_
    >>> import os
    >>> os.remove("config.json")
    """

    def save_json(self, path: str):
        with open(path, "w") as f:
            dict_ = dataclasses.asdict(self)
            json.dump(dict_, f, indent=1)

    @staticmethod
    def from_dict(dataclass: Type[Dataclass], d: Dict[str, Any]) -> Dataclass:
        for field in dataclasses.fields(dataclass):
            if dataclasses.is_dataclass(field.type):
                # nested dataclass:
                args_dict = d[field.name]
                nested_instance = JsonSerializable.from_dict(field.type, args_dict)
                d[field.name] = nested_instance
        return dataclass(**d) # type: ignore

    @classmethod
    def load_json(cls, path: str):
        with open(path) as f:
            args_dict = json.load(f)
        return JsonSerializable.from_dict(cls, args_dict)



if __name__ == "__main__":
    import doctest
    doctest.testmod()




import argparse
from typing import Optional, TypeVar, Type, List, Dict, Tuple, Any, Callable
import shlex
import pytest
import simple_parsing
from simple_parsing import InconsistentArgumentError, ArgumentParser, Formatter, ConflictResolution
from simple_parsing.wrappers import DataclassWrapper

from simple_parsing.utils import camel_case

xfail = pytest.mark.xfail
parametrize = pytest.mark.parametrize

def xfail_param(*args, reason:str):
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))

Dataclass = TypeVar("Dataclass")

class TestSetup():
    @classmethod
    def setup(cls: Type[Dataclass],
              arguments: Optional[str] = "",
              dest: Optional[str] = None,
              conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO,
              is_only_argparse_argument=True,              
              ) -> Dataclass:
        """Basic setup for a test.
        
        Keyword Arguments:
            arguments {Optional[str]} -- The arguments to pass to the parser (default: {""})
            dest {Optional[str]} -- the attribute where the argument should be stored. (default: {None})
        
        Returns:
            {cls}} -- the class's type.
        """
        parser = simple_parsing.ArgumentParser(conflict_resolution=conflict_resolution_mode)
        if dest is None:
            dest = camel_case(cls.__name__)
        
        parser.add_arguments(cls, dest=dest)

        if arguments is None:
            args = parser.parse_args()
        else:
            splits = shlex.split(arguments)
            args = parser.parse_args(splits)
        assert hasattr(args, dest), f"attribute '{dest}' not found in args {args}"
        instance: Dataclass = getattr(args, dest) #type: ignore
        delattr(args, dest)
        assert args == argparse.Namespace(), f"Namespace has leftover garbage values: {args}"
        return instance
    
    @classmethod
    def setup_multiple(cls: Type[Dataclass], num_to_parse: int, arguments: Optional[str] = "") -> Tuple[Dataclass, ...]:
        conflict_resolution_mode: ConflictResolution = ConflictResolution.ALWAYS_MERGE

        parser = simple_parsing.ArgumentParser(conflict_resolution=conflict_resolution_mode)
        class_name = camel_case(cls.__name__)
        for i in range(num_to_parse):
            parser.add_arguments(cls, f"{class_name}_{i}")

        if arguments is None:
            args = parser.parse_args()
        else:
            splits = shlex.split(arguments)
            args = parser.parse_args(splits)

        return tuple(getattr(args, f"{class_name}_{i}") for i in range(num_to_parse))
        

    @classmethod
    def get_help_text(cls, multiple=False, conflict_resolution_mode: ConflictResolution = ConflictResolution.AUTO) -> str:
        import contextlib
        from io import StringIO
        f = StringIO()
        with contextlib.suppress(SystemExit), contextlib.redirect_stdout(f):
            _ = cls.setup("--help", conflict_resolution_mode=conflict_resolution_mode)
        s = f.getvalue()
        return s


ListFormattingFunction = Callable[[List[Any]], str]
ListOfListsFormattingFunction = Callable[[List[List[Any]]], str]


def format_list_using_spaces(value_list: List[Any])-> str:
    return " ".join(str(p) for p in value_list)


def format_list_using_brackets(value_list: List[Any])-> str:
    return f"[{','.join(str(p) for p in value_list)}]"


def format_list_using_single_quotes(value_list: List[Any])-> str:
    return f"'{format_list_using_spaces(value_list)}'"


def format_list_using_double_quotes(value_list: List[Any])-> str:
    return f'"{format_list_using_spaces(value_list)}"'


def format_lists_using_brackets(list_of_lists: List[List[Any]])-> str:
    return " ".join(
        format_list_using_brackets(value_list)
        for value_list in list_of_lists
    )


def format_lists_using_double_quotes(list_of_lists: List[List[Any]])-> str:
    return " ".join(
        format_list_using_double_quotes(value_list)
        for value_list in list_of_lists
    )


def format_lists_using_single_quotes(list_of_lists: List[List[Any]])-> str:
    return " ".join(
        format_list_using_single_quotes(value_list)
        for value_list in list_of_lists
    )





from contextlib import suppress
from argparse import ArgumentError
from dataclasses import dataclass, field
from typing import *

import pytest

import simple_parsing
from simple_parsing import ArgumentParser, InconsistentArgumentError

from .testutils import *


@dataclass
class ContainerClass(TestSetup):
    a: Tuple[int]
    b: List[int]
    c: Tuple[str] = field(default_factory=tuple)
    d: List[int] = field(default_factory=list)



def test_single_element_list():
    container = ContainerClass.setup("--a 1 --b 4 --c 7 --d 10")
    assert container.a == (1,)
    assert container.b == [4]
    assert container.c == ('7',)
    assert container.d == [10]


def test_required_attributes_works():
    args = None
    with suppress(SystemExit), pytest.raises(ArgumentError):
        args = ContainerClass.setup("--b 4")
    assert args is None

    with suppress(SystemExit), pytest.raises(ArgumentError):
        args = ContainerClass.setup("--a 4")
    assert args is None

    args = ContainerClass.setup("--a 4 --b 5")
    assert args


def test_default_value():
    container = ContainerClass.setup("--a 1 2 3 --b 4 5 6")
    assert container.a == (1, 2, 3)
    assert container.b == [4, 5, 6]
    assert container.c == tuple()
    assert container.d == list()


@parametrize(
    "list_formatting_function",
    [
        format_list_using_spaces,
        # format_list_using_brackets,
        xfail_param(format_list_using_brackets, reason="TODO: decide which syntax we want to allow for single lists."),
        xfail_param(format_list_using_double_quotes, reason="TODO: decide which syntax we want to allow for single lists."),
        xfail_param(format_list_using_single_quotes, reason="TODO: decide which syntax we want to allow for single lists."),
    ])
@parametrize(
    "item_type, passed_values",
    [
        (int,   [1, 2]),
        (float, [1.1, 2.1]),
        (str,   ["a", "b"]),
        # (bool,  [True, True]),
    ]
)
def test_list_supported_formats(
        list_formatting_function: ListFormattingFunction,
        item_type: Type,
        passed_values: List[Any],
    ):
    @dataclass
    class SomeClass(TestSetup):
        a: List[item_type] = field(default_factory=list)  # type: ignore
        """some docstring for attribute 'a'"""

    arguments = "--a " + list_formatting_function(passed_values)
    print(arguments)

    some_class = SomeClass.setup(arguments)

    assert some_class.a == passed_values
    assert isinstance(some_class, SomeClass)
    assert len(some_class.a) == 2
    assert all(
        isinstance(v, item_type) for v in some_class.a
    )


@parametrize(
    "list_of_lists_formatting_function",
    [
        format_lists_using_brackets,
        format_lists_using_single_quotes,
        format_lists_using_double_quotes,
    ])
@parametrize(
    "item_type, passed_values",
    [
        (int,   [[1, 2], [4, 5], [7, 8]]),
        (float, [[1.1, 2.1], [4.2, 5.2], [7.2, 8.2]]),
        (str,   [["a", "b"], ["c", "d"], ["e", "f"]]),
        # (bool,  [[True, True], [True, False], [False, True]]),
    ]
)
def test_parse_multiple_with_list_attributes(
        list_of_lists_formatting_function: ListOfListsFormattingFunction,
        item_type: Type,
        passed_values: List[List[Any]],
    ):
    @dataclass
    class SomeClass(TestSetup):
        a: List[item_type] = field(default_factory=list)  # type: ignore
        """some docstring for attribute 'a'"""

    arguments = "--a " + list_of_lists_formatting_function(passed_values)
    print(arguments)

    classes = SomeClass.setup_multiple(3, arguments)

    assert len(classes) == 3
    for i, c_i in enumerate(classes):
        assert c_i.a == passed_values[i]
        assert isinstance(c_i, SomeClass)
        assert len(c_i.a) == 2
        assert all(
            isinstance(v, item_type) for v in c_i.a
        )




import argparse
import dataclasses
import shlex
from dataclasses import dataclass, field
from typing import *

import pytest

from .testutils import *
from simple_parsing import utils

@dataclass
class SomeDataclass:
    x: float = 123


@parametrize("t", [
    Tuple[int, ...],
    Tuple[str],
    Tuple,
    tuple,
])
def test_is_tuple(t: Type):
    assert utils.is_tuple(t)
    assert not utils.is_list(t)

@parametrize("t", [
    List[int],
    List[str],
    List,
    list,
    List[SomeDataclass],
])
def test_is_list(t: Type):
    assert utils.is_list(t)
    assert not utils.is_tuple(t)

@parametrize("t", [
    List[SomeDataclass],
    Tuple[SomeDataclass],
])
def test_is_list_of_dataclasses(t: Type):
    assert utils.is_tuple_or_list_of_dataclasses(t)

import enum
class Color(enum.Enum):
    RED = "RED"
    ORANGE = "ORANGE"
    BLUE = "BLUE"

class Temperature(enum.IntEnum):
    HOT = 1
    WARM = 0
    COLD = -1
    MONTREAL = -35

@parametrize("t", [
    Color,
    Temperature,
])
def test_is_enum(t: Type):
    assert utils.is_enum(t)


def test_json_serializable(datascience_example, tmpdir):
    HyperParameters, _ = datascience_example
    hparams = HyperParameters()
    import os
    filename = "hparams.json"
    hparams.save_json(os.path.join(tmpdir,  filename))
    hparams_ = HyperParameters.load_json(os.path.join(tmpdir, filename))
    assert hparams == hparams_



import shlex
from dataclasses import dataclass

import pytest

from simple_parsing import ArgumentParser

from .testutils import *


@dataclass
class Base(TestSetup):
    """ Some extension of base-class `Base` """
    common_attribute: int = 1


@dataclass
class ExtendedA(Base):
    a: int = 2

@dataclass
class ExtendedB(Base):
    b: int = 3


def inheritance_setup(arguments=""):
    parser = ArgumentParser()
    parser.add_arguments(ExtendedA, "extended_a")
    parser.add_arguments(ExtendedB, "extended_b")

    splits = shlex.split(arguments)
    args = parser.parse_args(splits)

    exta = args.extended_a
    extb = args.extended_b
    return exta, extb

def test_simple_subclassing_no_args():
    extended = ExtendedA.setup()
    assert extended.a == 2
    assert extended.common_attribute == 1


def test_simple_subclassing_with_args():
    extended = ExtendedA.setup("--common_attribute 123 --a 56")
    assert extended.a == 56
    assert extended.common_attribute == 123
    

@xfail(reason="TODO: make sure this is how people would want to use this feature.")
def test_subclasses_with_same_base_class_no_args():
    ext_a, ext_b = inheritance_setup()
    
    assert ext_a.common_attribute == 1
    assert ext_a.a == 2

    assert ext_b.common_attribute == 1
    assert ext_b.b == 3


@xfail(reason="TODO: make sure this is how people would want to use this feature.")
def test_subclasses_with_same_base_class_with_args():
    ext_a, ext_b = inheritance_setup("--a 10 --b 20 --a 30 --c 40")
    
    assert ext_a.common_attribute == 10
    assert ext_a.a == 20

    assert ext_b.common_attribute == 30
    assert ext_b.b == 40


import argparse
import dataclasses
import shlex
from dataclasses import dataclass, field
from typing import *

import pytest

from simple_parsing import (InconsistentArgumentError,
                            ArgumentParser)

from .testutils import *

@parametrize("num_instances", [1, 2, 5])
@parametrize(
    "some_type, default_value",
    [
        (int,   123524),
        (float, 123.456),
        (str,   "bob"),
        (bool,  True),
    ]
)
def test_parse_multiple_with_no_arguments_sets_default_value(num_instances: int, some_type: Type, default_value: Any):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type = default_value  # type: ignore
        """some docstring for attribute 'a'"""

    classes = SomeClass.setup_multiple(num_instances, "")
    assert len(classes) == num_instances
    for i in range(num_instances):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == default_value
        assert isinstance(c_i.a, some_type)


@parametrize("num_instances", [1, 2, 5])
@parametrize(
    "some_type, default_value,  passed_value",
    [
        (int,   123524,     12),
        (float, 123.456,    -12.3),
        (str,   "bob",      "random"),
        (bool,  True,       False),
    ])
def test_parse_multiple_with_single_arg_value_sets_that_value_for_all_instances(
        num_instances: int,
        some_type: Type,
        default_value: Any,
        passed_value: Any
    ):

    @dataclass
    class SomeClass(TestSetup):
        a: some_type = default_value  # type: ignore
        """some docstring for attribute 'a'"""
    classes = SomeClass.setup_multiple(num_instances, f"--a {passed_value}")

    assert len(classes) == num_instances
    for i in range(num_instances):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == passed_value
        assert isinstance(c_i.a, some_type)


@parametrize(
    "some_type, default_value,  passed_values",
    [
        (int,   123524,     [1, 2, 3]),
        (float, 123.456,    [4.5, -12.3, 9]),
        (str,   "bob",      ["random", "triceratops", "cocobongo"]),
        (bool,  True,       [False, True, False]),
    ])
def test_parse_multiple_with_provided_value_for_each_instance(
        some_type: Type,
        default_value: Any,
        passed_values: List[Any]
    ):

    @dataclass
    class SomeClass(TestSetup):
        a: some_type = default_value  # type: ignore
        """some docstring for attribute 'a'"""
    # TODO: maybe test out other syntaxes for passing in multiple argument values? (This looks a lot like passing in a list of values..)
    arguments = f"--a {' '.join(str(p) for p in passed_values)}"
    classes = SomeClass.setup_multiple(3, arguments)

    assert len(classes) == 3
    for i in range(3):
        c_i = classes[i]
        assert isinstance(c_i, SomeClass)
        assert c_i.a == passed_values[i]
        assert isinstance(c_i.a, some_type)


@parametrize("some_type", [int, float, str, bool])        
def test_parse_multiple_without_required_arguments(some_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: some_type # type: ignore
        """some docstring for attribute 'a'"""

    with pytest.raises(SystemExit):
        some_class = SomeClass.setup_multiple(2, "")

@parametrize("container_type", [List, Tuple])
@parametrize("item_type", [int, float, str, bool])
def test_parse_multiple_without_required_container_arguments(container_type: Type, item_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: container_type[item_type] # type: ignore
        """some docstring for attribute 'a'"""

    with pytest.raises(SystemExit):
        _ = SomeClass.setup_multiple(3, "")


@parametrize("container_type", [List, Tuple])
@parametrize("item_type", [int, float, str, bool])
def test_parse_multiple_with_arg_name_without_arg_value(container_type: Type, item_type: Type):
    @dataclass
    class SomeClass(TestSetup):
        a: container_type[item_type] # type: ignore
        """some docstring for attribute 'a'"""

    with pytest.raises(SystemExit):
        _ = SomeClass.setup_multiple(3, "--a")





"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
import argparse
import collections
import dataclasses
import enum
import inspect
import logging
import re
import textwrap
import typing
import warnings
from collections import defaultdict, namedtuple
from typing import *


# logging.basicConfig(level=logging.WARN)
# logger = logging.getLogger("simple_parsing")
logger = logging.getLogger(__name__)

from . import docstring, utils
from .wrappers import DataclassWrapper, FieldWrapper
from .utils import Dataclass, DataclassType, MutableField

Conflict = Tuple[DataclassType, str, List[DataclassWrapper]]

class ConflictResolution(enum.Enum):
    """Used to determine which action to take when adding arguments for the same dataclass in two different destinations.
    
    - NONE: Dissallow using the same dataclass in two different destinations without explicitly setting a distinct prefix for at least one of them.
    - EXPLICIT: When adding arguments for a dataclass that is already present, the argparse arguments for each class will use their full absolute path as a prefix.
    - ALWAYS_MERGE: When adding arguments for a dataclass that is already present, the arguments for the first and second destinations will be set using the same name, 
        and the values for each will correspond to the first and second passed values, respectively.
        This will change the argparse type for that argument into a list of the original item type.
    - AUTO: TODO:
    
    """
    NONE = -1
    EXPLICIT = 0
    ALWAYS_MERGE = 1
    AUTO = 2

class ArgumentParser(argparse.ArgumentParser):
    def __init__(self, conflict_resolution: ConflictResolution = ConflictResolution.AUTO, *args, **kwargs):
        if "formatter_class" not in kwargs:
            kwargs["formatter_class"] = utils.Formatter
        super().__init__(*args, **kwargs)
        # the kind of prefixing mechanism to use.
        self.conflict_resolution = conflict_resolution

        # two-level dictionary that maps from (dataclass, string_prefix) -> DataclassWrapper. 
        self._wrappers: Dict[DataclassType, Dict[str, List[DataclassWrapper]]] = defaultdict(lambda: defaultdict(list))

        self._fixed_wrappers: Dict[DataclassType, Dict[str, DataclassWrapper]]

        self.constructor_arguments: Dict[str, Dict[str, Any]] = defaultdict(dict)

        # # a set to check which arguments have been added so far.
        # self.added_arguments: Set[str] = set()

    def add_arguments(self, dataclass: DataclassType, dest: str, prefix=""):
        """Adds corresponding command-line arguments for this class to the parser.
        
        Arguments:
            dataclass {DataclassType} -- The dataclass for which to add fields as arguments in the parser
        
        Keyword Arguments:
            dest {str} -- The destination attribute of the `argparse.Namespace` where the dataclass instance will be stored after calling `parse_args()`
            prefix {str} -- An optional prefix to add prepend to the names of the argparse arguments which will be generated for this dataclass.
            This can be useful when registering multiple distinct instances of the same dataclass.

        """
        for prefix, wrappers in self._wrappers[dataclass].items():
            destinations = [wrapper.dest for wrapper in wrappers]
            if dest in destinations:
                self.error(textwrap.dedent(f"""\
                    Destination attribute {dest} is already used for dataclass of type {dataclass}.
                    Make sure all destinations are unique.
                    """))
        new_wrapper: DataclassWrapper[DataclassType] = DataclassWrapper(dataclass, dest)
        new_wrapper.prefix = prefix
        wrapper = self._register_dataclass(new_wrapper)
        logger.debug(f"added wrapper:\n{wrapper}\n")

    def parse_known_args(self, args=None, namespace=None):
        # NOTE: since the usual ArgumentParser.parse_args() calls parse_known_args, we therefore just need to overload the parse_known_args method.
        self._preprocessing()
        parsed_args, unparsed_args = super().parse_known_args(args, namespace)
        return self._postprocessing(parsed_args), unparsed_args

    def _register_dataclass(self, new_wrapper: DataclassWrapper[DataclassType]) -> DataclassWrapper[DataclassType]:
        """registers the given dataclass to be parsed later.
        
        Arguments:
            dataclass {Type[Dataclass]} -- The dataclass to register
            dest {str} -- a string which is to be used to  NamedTuple used to keep track of where to store the resulting instance and the number of instances.
        """
        logger.debug(f"Registering new DataclassWrapper: {new_wrapper}")
        self._wrappers[new_wrapper.dataclass][new_wrapper.prefix].append(new_wrapper)        
        for child in new_wrapper.descendants:
            self._wrappers[child.dataclass][child.prefix].append(child)
        return new_wrapper

    def _unregister_dataclass(self, wrapper: DataclassWrapper[DataclassType]):
        logger.debug(f"Unregistering DataclassWrapper {wrapper}")
        self._remove(wrapper)
        for child in wrapper.descendants:
            logger.debug(f"\tAlso Unregistering Child DataclassWrapper {child}")
            self._remove(child)
    
    def _remove(self, wrapper: DataclassWrapper):
        self._wrappers[wrapper.dataclass][wrapper.prefix].remove(wrapper)
        if len(self._wrappers[wrapper.dataclass][wrapper.prefix]) == 0:
            self._wrappers[wrapper.dataclass].pop(wrapper.prefix)
    
    def _preprocessing(self):
        logger.debug("\nPREPROCESSING\n")

        self._fixed_wrappers = self._fix_conflicts()
        # logger.debug(f"Fixed wrappers: {self._fixed_wrappers}")
        # Create one argument group per dataclass type
        for dataclass in self._fixed_wrappers:
            for prefix, wrapper in self._fixed_wrappers[dataclass].items():
                logger.debug(f"Adding arguments for dataclass: {dataclass} at destinations {wrapper.destinations}, multiple={wrapper.multiple}, prefix = '{wrapper.prefix}'")                
                wrapper.add_arguments(parser=self)


    def _postprocessing(self, parsed_args: argparse.Namespace) -> argparse.Namespace:
        """Instantiate the dataclasses from the parsed arguments and add them to their destination key in the namespace
        
        Arguments:
            parsed_args {argparse.Namespace} -- the result of calling super().parse_args()
        
        Returns:
            argparse.Namespace -- The namespace, with the added attributes for each dataclass.
            TODO: Try and maybe return a nicer, typed version of parsed_args (a Namespace subclass maybe?)  
        """
        logger.debug("\nPOST PROCESSING\n")
        logger.debug(f"parsed args: {parsed_args}")
        
        # create the constructor arguments for each instance by consuming all the attributes from `parsed_args` 
        parsed_args = self._consume_constructor_arguments(parsed_args)
        logger.debug(f"leftover arguments: {parsed_args}")
        
        logger.debug(f"Constructor arguments:")
        for key, args_dict in self.constructor_arguments.items():
            logger.debug(f"\t{key}: {args_dict}")

        self._set_instances_in_args(parsed_args)

        logger.debug(f"Final parsed args:")
        for key, value in vars(parsed_args).items():
            logger.debug(f"\t{key}: {value}")
           
        return parsed_args

    def _set_instances_in_args(self, parsed_args: argparse.Namespace) -> argparse.Namespace:
        
        # we now have all the constructor arguments for each instance.
        # we can now sort out the different dependencies, and create the instances.
        wrappers: Dict[str, DataclassWrapper] = self._wrapper_for_every_destination
        # we construct the 'tree' of dependencies from the bottom up,
        # starting with nodes that are children.
        destinations = self.constructor_arguments.keys()
        nesting_level = lambda destination_attribute: destination_attribute.count(".")

        for destination, wrapper in sorted(wrappers.items(), key=lambda k_v: k_v[1].nesting_level, reverse=True):
            logger.debug(f"wrapper name: {wrapper.attribute_name}, destination: {destination}")
            constructor = wrapper.instantiate
            constructor_args = self.constructor_arguments[destination]
            # create the dataclass instance.
            instance = constructor(constructor_args)
            
            if wrapper._parent is not None:
                parts = destination.split(".")
                parent = ".".join(parts[:-1])
                attribute_in_parent = parts[-1]
                # if this instance is an attribute in another dataclass,
                # we set the value in the parent's constructor arguments
                # at the associated attribute to this instance.
                self.constructor_arguments[parent][attribute_in_parent] = instance
                # self.constructor_arguments.pop(destination) # remove the 'args dict' for this child class.
                logger.debug(f"Setting a value at attribute {attribute_in_parent} in parent {parent}.")
            else:
                # if this destination is a top-level attribute, we set the attribute
                # on the returned parsed_args.
                logger.debug(f"setting attribute '{destination}' on the parsed_args to a value of {instance}")
                assert not hasattr(parsed_args, destination), "Namespace should not already have a '{destination}' attribute! (namespace: {parsed_args}) "
                setattr(parsed_args, destination, instance)
        return parsed_args

    def _consume_constructor_arguments(self, parsed_args: argparse.Namespace) -> argparse.Namespace:
        """Create the constructor arguments for each instance by consuming all the attributes from `parsed_args` 
        
        Args:
            parsed_args (argparse.Namespace): the argparse.Namespace returned from super().parse_args().
        
        Returns:
            argparse.Namespace: The namespace, without the consumed arguments.
        """
        wrappers: Dict[str, DataclassWrapper] = self._wrapper_for_every_destination
        parsed_arg_values = vars(parsed_args)
        # TODO: it would be cleaner if the CustomAction was working!
        # Here we imitate a custom action, by having the FieldWrappers be callables
        for wrapper in wrappers.values():
            for field in wrapper.fields:
                if not field.field.init:
                    continue
                values = parsed_arg_values.get(field.dest, field.defaults)
                # call the action manually.
                # this sets the right value in the `self.constructor_arguments` dictionary.
                field(parser=self, namespace=parsed_args, values=values, option_string=None)
                
        #Clean up the 'parsed_args' by deleting all the consumed attributes.
        deleted_values: Dict[str, Any] = {}
        for wrapper in wrappers.values():
            for field in wrapper.fields:
                deleted_values[field.dest] = parsed_arg_values.pop(field.dest, None)
        leftover_args = argparse.Namespace(**parsed_arg_values)
        logger.debug(f"deleted values: {deleted_values}")
        logger.debug(f"leftover args: {leftover_args}")
        return leftover_args

    def _get_conflicting_group(self, all_wrappers: Dict[DataclassType, Dict[str, List[DataclassWrapper[DataclassType]]]]) -> Optional[Conflict]:
        """Return the dataclass, prefix, and conflicing DataclassWrappers.
        """
        for dataclass in all_wrappers.keys():
            for prefix in all_wrappers[dataclass].keys():
                wrappers = all_wrappers[dataclass][prefix].copy()
                if len(wrappers) >= 2:
                    return dataclass, prefix, wrappers
        return None


    def _conflict_exists(self, all_wrappers: Dict[DataclassType, Dict[str, List[DataclassWrapper[DataclassType]]]]) -> bool:
        """Return True whenever a conflict exists (multiple DataclassWrappers share the same dataclass and prefix."""
        for dataclass in all_wrappers.keys():
            for prefix, wrappers in all_wrappers[dataclass].items():
                assert all(wrapper.prefix == prefix for wrapper in wrappers), "Misplaced DataclassWrapper!"
                if len(wrappers) >= 2:
                    return True
        return False

    def _fix_conflicts(self) ->  Dict[DataclassType, Dict[str, DataclassWrapper]]:
        while self._conflict_exists(self._wrappers):
            conflict = self._get_conflicting_group(self._wrappers)
            assert conflict is not None
            dataclass, prefix, wrappers = conflict
            logger.info(f"The following {len(wrappers)} wrappers are in conflict, as they share the same dataclass and prefix:\n" + "\n".join(str(w) for w in wrappers))
            logger.debug(f"(Conflict Resolution mode is {self.conflict_resolution})")
            if self.conflict_resolution == ConflictResolution.NONE:           
                self.error(
                    "The following wrappers are in conflict, as they share the same dataclass and prefix:\n" +
                    "\n".join(str(w) for w in wrappers) +
                    f"(Conflict Resolution mode is {self.conflict_resolution})"
                )
                    
            elif self.conflict_resolution == ConflictResolution.EXPLICIT:
                self._fix_conflict_explicit(conflict)

            elif self.conflict_resolution == ConflictResolution.ALWAYS_MERGE:
                self._fix_conflict_merge(conflict)

            elif self.conflict_resolution == ConflictResolution.AUTO:
                self._fix_conflict_auto(conflict)

        assert not self._conflict_exists(self._wrappers)
        fixed_wrappers: Dict[DataclassType, Dict[str, DataclassWrapper[DataclassType]]] = defaultdict(dict)
        for dataclass in self._wrappers:
            for prefix, wrappers in self._wrappers[dataclass].items():
                assert len(wrappers) == 1
                wrapper = wrappers[0]
                assert wrapper.prefix == prefix 
                fixed_wrappers[dataclass][prefix] = wrapper 
        return fixed_wrappers

    def _fix_conflict_explicit(self, conflict):
        # logger.debug("fixing explicit conflict: ", conflict)
        dataclass, prefix, wrappers = conflict
        assert prefix == "", "Wrappers for the same dataclass can't have the same user-set prefix when in EXPLICIT mode!"
        # remove all wrappers for that prefix
        for wrapper in wrappers:
            self._unregister_dataclass(wrapper)
            wrapper.explicit = True
            self._register_dataclass(wrapper)
        assert not self._wrappers[dataclass][prefix], self._wrappers[dataclass][prefix]
        # remove the prefix from the dict so we don't have to deal with empty lists.
        self._wrappers[dataclass].pop(prefix)

    def _fix_conflict_auto(self, conflict):
        # logger.debug("fixing conflict: ", conflict)
        dataclass, prefix, wrappers = conflict
        prefixes: List[List[str]] = [wrapper.dest.split(".") for wrapper in wrappers]
        # IDEA:
        # while the prefixes are the same, starting from the left, remove the first word.
        # Stop when they become different.


        sentences: Dict[int, List[str]] = {
            i: sentence for i, sentence in enumerate(prefixes)   
        }
        index_prefixes = {
            i: "" for i, sentence in enumerate(prefixes)
        }
        logger.debug(f"conflicting prefixes: {prefixes}")
        # TODO: do something here with the prefixes.
        def differentiate(prefixes: List[List[str]]):
            result: Dict[str, List[List[str]]] = defaultdict(list)
            for sentence in prefixes:
                first_word = sentence[0]
                result[first_word].append(sentence)

            return_dict = {}
            for first_word, sentences in result.items():
                if len(sentences) == 1:
                    return_dict[first_word] = ".".join(sentences[0])
                else:
                    sentences_without_first_word = [sentence[1:] for sentence in sentences]
                    return_dict[first_word] = differentiate(sentences_without_first_word)
            return return_dict

        prefix_dict = differentiate(prefixes)
        logger.debug(f"Prefix dict: {prefix_dict}")
        dest_to_wrapper: Dict[str, DataclassWrapper] = { wrapper.dest: wrapper for wrapper in wrappers }
        prefix_to_wrapper: Dict[str, DataclassWrapper] = {}
        for dest, wrapper in dest_to_wrapper.items():
            parts = dest.split(".")
            _prefix_dict = prefix_dict
            while len(parts) > 1:
                _prefix_dict = _prefix_dict[parts[0]]
                parts = parts[1:]
            prefix = _prefix_dict[parts[0]]
            prefix_to_wrapper[prefix] = wrapper
        
        for prefix, wrapper in prefix_to_wrapper.items():
            logger.debug(f"Wrapper for attribute {wrapper.dest} has prefix '{prefix}'.")
            self._unregister_dataclass(wrapper)
            wrapper.prefix = prefix + "."
            self._register_dataclass(wrapper)


    def _fix_conflict_merge(self, conflict):
        """Fix conflicts using the merging approach:
        The first wrapper is kept, and the rest of the wrappers are absorbed into the first wrapper.

        # TODO: check that the ordering of arguments is still preserved!
        
        Args:
            conflict ([type]): [description]
        """
        dataclass, prefix, wrappers = conflict
        assert len(wrappers) > 1
        first_wrapper: DataclassWrapper = None
        for wrapper in wrappers:
            self._unregister_dataclass(wrapper)
            if first_wrapper is None:
                first_wrapper = wrapper
            else:
                first_wrapper.merge(wrapper)
        
        assert first_wrapper.multiple
        self._register_dataclass(first_wrapper)

    
    @property
    def _wrapper_for_every_destination(self) -> Dict[str, DataclassWrapper[DataclassType]]:
        """Returns a dictionary where for every key (a destination), we return the associated DataclassWrapper.
        NOTE: multiple destinations can share the same DataclassWrapper instance.
        
        Returns:
            Dict[str, DataclassWrapper[Dataclass]] -- [description]
        """
        wrapper_for_destination: Dict[str, DataclassWrapper[DataclassType]] = {}
        for dataclass in self._fixed_wrappers.keys():
            for prefix, wrapper in self._fixed_wrappers[dataclass].items():
                logger.debug(f"prefix: '{prefix}', wrapper destinations: {wrapper.destinations}")
                for dest in wrapper.destinations:
                    assert dest not in wrapper_for_destination, f"There is already a wrapper for destination '{dest}': {wrapper_for_destination[dest]}"
                    wrapper_for_destination[dest] = wrapper
                    logger.debug(f"Setting a wrapper for destination {dest}")
        return wrapper_for_destination
        


"""Simple, Elegant Argument parsing.
@author: Fabrice Normandin
"""
from .utils import Formatter, InconsistentArgumentError, MutableField
from .parsing import ArgumentParser, ConflictResolution

__all__ = [
    "ArgumentParser",
    "InconsistentArgumentError",
    "Formatter",
    "ConflictResolution",
    "MutableField"
]


import setuptools

with open("README_pipy.md", "r") as fh:
    long_description = fh.read()
print("PACKAGES FOUND:", setuptools.find_packages())

setuptools.setup(
    name="simple_parsing",
    version="0.0.3post2",
    author="Fabrice Normandin",
    author_email="fabrice.normandin@gmail.com",
    description="A small utility for simplifying and cleaning up argument parsing scripts.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lebrice/SimpleParsing",
    packages=["simple_parsing"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)